BASE_URL = "https://app.lila.dev"
WIDTH = 1536
HEIGHT = 1152

MAX_LOGS_DISPLAY = 3

API_KEY_URL = "https://app.lila.dev/manage-api-keys"

INCLUDE_ATTRIBUTES = [
    "title",
    "type",
    "name",
    "role",
    "tabindex",
    "aria-label",
    "placeholder",
    "value",
    "alt",
    "aria-expanded",
]

MAX_ATTEMPTS = 30
